import unittest
from .test_base_model import TestBaseModel
from .test_storable_model import TestStorableModel
from .test_submodel import TestSubmodel
from .test_validators import TestValidators
from .test_reference import TestReference
from .test_config import TestConfig


if __name__ == "__main__":
    unittest.main()
